module MetaRPC
  # Module used to describe callable JSON RPC methods
  module Callable
    def self.included(klass)
      super
      klass.send(:extend, ClassMethods)
    end

    def raise_json_rpc_error(code)
      raise RPCError.new(code), "Failed to call JSON RPC method: #{code}"
    end

    DYNAMIC_TYPE_SUFFIXES = {
      nil => :not_nullable_nor_spread,
      '?*' => :nullable_spread,
      '*?' => :spread_nullable,
      '?' => :nullable,
      '*' => :spread
    }.freeze
    DYNAMIC_TYPE_REGEXP = /^(\w+)(\?|\?\*|\*\?|\*)?$/.freeze

    def validate_type(item, klass, nullable: false)
      return raise_json_rpc_error(:invalid_params) if item.nil? && !nullable

      raise_json_rpc_error(:invalid_params) unless (item.present? && item.is_a?(klass)) || item.nil?
    end

    def validate_spread(items, klass, items_nullable: false, spread_nullable: false)
      raise_json_rpc_error(:invalid_params) if items.nil? && !spread_nullable
      items&.each { |item| validate_type(item, klass, nullable: items_nullable) }
    end

    def validate_dynamic_type(item, contract_type)
      matched_type = contract_type.match(DYNAMIC_TYPE_REGEXP)
      klass = matched_type[1].constantize
      type_suffix = DYNAMIC_TYPE_SUFFIXES[matched_type[2]]

      case type_suffix
      when :not_nullable_nor_spread
        validate_type(item, klass)
      when :nullable
        validate_type(item, klass, nullable: true)
      when :spread
        validate_spread(item, klass)
      when :nullable_spread
        validate_spread(item, klass, items_nullable: true)
      when :spread_nullable
        validate_spread(item, klass, spread_nullable: true)
      else
        raise_json_rpc_error(:internal_error)
      end
    end

    def validate_item(item, contract_type)
      case contract_type
      when Hash
        validate_hash(item, contract_type)
      when Array
        validate_array(item, contract_type)
      when Class
        raise_json_rpc_error(:invalid_params) unless item.is_a?(contract_type)
      when String
        validate_dynamic_type(item, contract_type)
      else
        raise_json_rpc_error(:internal_error)
      end
    rescue NameError
      raise_json_rpc_error(:internal_error)
    end

    def validate_array(args, contract)
      contract.each_with_index do |contract_type, i|
        validate_item(args[i], contract_type)
      end
    end

    def validate_hash(args, contract)
      contract.each do |arg_name, contract_type|
        raise_json_rpc_error(:invalid_params) unless args.key?(arg_name)

        validate_item(args[arg_name], contract_type)
      end
    end

    # Injected class methods
    module ClassMethods
      def json_rpc_methods
        @__json_rpc_hooked_methods
      end

      def handle_json_callable(method_name)
        old_method = instance_method(method_name)
        return if (@__json_rpc_hooked_methods ||= {}).key?(method_name)

        @__ignoring_added_methods = true
        params_contract = @__json_rpc_method_params
        error_handler = @__on_json_rpc_error

        define_method method_name do |*args, &block|
          case params_contract
          when nil
            raise_json_rpc_error(:invalid_params) unless args.empty?
          when Array
            validate_item(args, params_contract)
          when Hash
            raise_json_rpc_error(:invalid_params) unless args[0].is_a?(Hash)
            validate_item(args[0], params_contract)
          else
            raise_json_rpc_error(:internal_error)
          end

          old_method.bind(self).call(*args, &block)
        rescue RPCError => err
          raise err

        rescue StandardError => err
          error_handler&.call(err)
          raise_json_rpc_error(:server_error)
        end

        @__ignoring_added_methods = false
        @__json_rpc_hooked_methods[method_name] = {
          description: @__json_rpc_method_description,
          params: @__json_rpc_method_params
        }
      end

      def on_json_rpc_error(&func)
        @__on_json_rpc_error = func
      end

      def json_rpc_method(description, params: nil)
        @__next_method_is_json_rpc = true
        @__json_rpc_method_description = description
        @__json_rpc_method_params = params
      end

      def method_added(method_name)
        super
        handle_json_callable(method_name) if @__next_method_is_json_rpc && !@__ignoring_added_methods
        @__next_method_is_json_rpc = false
      end
    end
  end
end
