module MetaRPC
  class Client
    attr_reader :metarpc_url, :metarpc_pending_batch, :metarpc_header_setter_func

    def initialize(url, header_setter_func = nil)
      @metarpc_url = URI.parse(url)
      @metarpc_header_setter_func = header_setter_func
    end

    def respond_to_missing?(_sym)
      true
    end

    def method_missing(sym, *args, &blk)
      call = metarpc_build_method(sym, args)
      return metarpc_pending_batch << { call: call, on_response: blk } if metarpc_pending_batch

      metarpc_execute_request(call)
    end

    def rpc_batch(&blk)
      @metarpc_pending_batch = []
      instance_exec(&blk)
      metarpc_execute_request(metarpc_pending_batch.map { |p| p[:call] })
      @metarpc_pending_batch = nil
    end

    def metarpc_build_method(method, args)
      {
        jsonrpc: '2.0',
        method: method,
        params: args[0].is_a?(Hash) && args.count == 1 ? args[0] : args,
        id: SecureRandom.uuid
      }
    end

    def metarpc_execute_request(call)
      body = call.to_json
      headers = metarpc_header_setter_func.call(body) if metarpc_header_setter_func.present?
      (headers ||= {})['Content-Type'] = 'text/json'

      http = Net::HTTP.new(metarpc_url.host, metarpc_url.port)
      request = Net::HTTP::Post.new(metarpc_url.request_uri, headers)
      request.body = body

      metarpc_parse_response(http.request(request).body)
    end

    def metarpc_parse_response(raw_response)
      response = JSON.parse(raw_response)
      return metarpc_handle_batch_response(response) if metarpc_pending_batch

      response.deep_symbolize_keys!
      raise RpcClientError.new(response[:error]), "#{response[:error][:code]} #{response[:error][:message]}" if response.key?(:error)

      response[:result]
    end

    def metarpc_handle_batch_response(response)
      response.each do |raw_call_response|
        call_response = raw_call_response.deep_symbolize_keys
        pending_batch = metarpc_pending_batch.find { |batch| batch[:call][:id] == call_response[:id] }
        next unless pending_batch[:on_response].present?

        response = ResponseWrapper.new(call_response)
        pending_batch[:on_response].call(response)
      end
    end
  end
end
